import platform
import sys
import logging
import argparse
import json
from datetime import datetime
from pathlib import Path
from socket import gaierror
import validators
from rich.console import Console
from rich.style import Style
from rich.logging import RichHandler
from rich.table import Table
from rich import box
from OpenSSL.crypto import FILETYPE_PEM, load_certificate
from tlstrust import __version__, TrustStore, trust_stores_from_chain
from tlstrust.util import get_certificate_chain, get_cn_or_org

__module__ = "tlstrust.cli"

assert sys.version_info >= (3, 9), "Requires Python 3.9 or newer"
logger = logging.getLogger(__name__)
console = Console()
CLI_COLOR_OK = "dark_sea_green2"
CLI_COLOR_NOK = "light_coral"
CLI_VALUE_TRUSTED = "Trusted"
CLI_VALUE_NOT_TRUSTED = "Not Trusted"
DEFAULT_PORT = 443


def styled_boolean(
    value: bool, colors: tuple[str, str] = (CLI_COLOR_OK, CLI_COLOR_NOK)
) -> str:
    if not isinstance(value, bool):
        raise TypeError(f"{type(value)} provided")
    color = colors[0] if value else colors[1]
    val = CLI_VALUE_TRUSTED if value else CLI_VALUE_NOT_TRUSTED
    with console.capture() as capture:
        console.print(val, style=Style(color=color))
    return capture.get().strip()


def date_diff(comparer: datetime) -> str:
    interval = comparer - datetime.utcnow()
    if interval.days < -1:
        return f"Expired {int(abs(interval.days))} days ago"
    if interval.days == -1:
        return "Expired yesterday"
    if interval.days == 1:
        return "Expires tomorrow"
    if interval.days > 365:
        return (
            f"Expires in {interval.days} days ({int(round(interval.days/365))} years)"
        )
    if interval.days > 1:
        return f"Expires in {interval.days} days"
    return "Expires today"


def output(store: TrustStore) -> Table:
    subject_common_name = get_cn_or_org(store.certificate)
    title = f'{"Trusted ✓✓✓" if store.is_trusted else "Not Trusted"}\nRoot Certificate {subject_common_name}\n{date_diff(store.certificate.to_cryptography().not_valid_after)}'
    caption = f"SKI {store.key_identifier}"
    title_style = Style(
        bold=True, color=CLI_COLOR_OK if store.is_trusted else CLI_COLOR_NOK
    )
    table = Table(title=title, caption=caption, title_style=title_style, box=box.SIMPLE)
    table.add_column(
        "Root Trust Store", justify="right", style="dark_turquoise", no_wrap=True
    )
    table.add_column("Result", justify="left", no_wrap=True)
    for name, is_trusted in store.all_results.items():
        table.add_row(name, styled_boolean(is_trusted))
    console.print(table)
    console.print()


def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "targets",
        nargs="*",
        help="All unnamed arguments are hosts (and ports) targets to test. ~$ tlstrust apple.com:443 github.io localhost:3000",
    )
    parser.add_argument(
        "-C",
        "--client-pem",
        help="path to PEM encoded client certificate, url or file path accepted",
        dest="client_pem",
        default=None,
    )
    parser.add_argument(
        "--disable-sni",
        help="Do not negotiate SNI using INDA encoded host",
        dest="disable_sni",
        action="store_true",
    )
    parser.add_argument(
        "-O",
        "--json-file",
        help="Store to file as JSON",
        dest="json_file",
        default=None,
    )
    parser.add_argument(
        "-v",
        "--errors-only",
        help="set logging level to ERROR (default CRITICAL)",
        dest="log_level_error",
        action="store_true",
    )
    parser.add_argument(
        "-vv",
        "--warning",
        help="set logging level to WARNING (default CRITICAL)",
        dest="log_level_warning",
        action="store_true",
    )
    parser.add_argument(
        "-vvv",
        "--info",
        help="set logging level to INFO (default CRITICAL)",
        dest="log_level_info",
        action="store_true",
    )
    parser.add_argument(
        "-vvvv",
        "--debug",
        help="set logging level to DEBUG (default CRITICAL)",
        dest="log_level_debug",
        action="store_true",
    )
    parser.add_argument("--version", dest="show_version", action="store_true")
    args = parser.parse_args()

    log_level = logging.CRITICAL
    if args.log_level_error:
        log_level = logging.ERROR
    if args.log_level_warning:
        log_level = logging.WARNING
    if args.log_level_info:
        log_level = logging.INFO
    if args.log_level_debug:
        log_level = logging.DEBUG
    handlers = []
    log_format = "%(asctime)s - %(name)s - [%(levelname)s] %(message)s"
    if sys.stdout.isatty():
        log_format = "%(message)s"
        handlers.append(RichHandler(rich_tracebacks=True))
    logging.basicConfig(format=log_format, level=log_level, handlers=handlers)

    def version():
        print(
            f"{__version__} Python {sys.version} {platform.platform()} {platform.uname().node} {platform.uname().release} {platform.version()}"
        )

    if args.show_version:
        version()
        sys.exit(0)
    if len(args.targets) == 0:
        parser.print_help(sys.stderr)
        sys.exit(1)

    if args.client_pem:
        client_certificate = load_certificate(
            FILETYPE_PEM, Path(args.client_pem).read_bytes()
        )
        logger.debug(
            f"client certificate issuer: {client_certificate.get_issuer().commonName}"
        )

    evaluation_start = datetime.utcnow()
    domains = []
    for target in args.targets:
        pieces = target.split(":")
        host, port = None, None
        if len(pieces) == 2:
            host, port = pieces
        if len(pieces) == 1:
            host = pieces[0]
            port = DEFAULT_PORT
        if validators.domain(host) is not True:
            raise AttributeError(f"host {host} is invalid")
        domains.append((host, int(port)))

    results = []
    for domain, port in domains:
        query = {
            "host_name": domain,
            "port_number": int(port),
            "use_sni": not args.disable_sni,
        }
        try:
            res = get_certificate_chain(domain, int(port), use_sni=not args.disable_sni)
            if not res:
                query["error"] = f"No supported TLS protocols {host}:{port}"
                results.append({"_query": query})
                console.print(query["error"])
                continue
            chain, peer_addr = res
        except (TimeoutError, ConnectionRefusedError, gaierror) as ex:
            query["error"] = f"{str(ex)} {host}:{port}"
            results.append({"_query": query})
            console.print(query["error"])
            continue
        query["peer_address"] = peer_addr
        console.print(f"{host}:{port} ({peer_addr})")
        for trust_store in trust_stores_from_chain(chain):
            data = trust_store.to_dict()
            data["_query"] = query
            results.append(data)
            if not args.json_file:
                output(trust_store)

    execution_duration_seconds = (datetime.utcnow() - evaluation_start).total_seconds()
    if args.json_file:
        json_data = {
            "generator": f"{__module__} {__version__}",
            "targets": [f"{domain}:{port}" for domain, port in domains],
            "execution_date": datetime.utcnow().replace(microsecond=0).isoformat(),
            "execution_duration_seconds": execution_duration_seconds,
            "evaluations": results,
        }
        json_path = Path(args.json_file)
        if json_path.is_file():
            json_path.unlink()
        json_path.write_text(
            json.dumps(json_data, sort_keys=True, default=str), encoding="utf8"
        )
        console.print(f"Saved to: {json_path.absolute()}")
    else:
        console.print(f"Evaluation duration seconds {execution_duration_seconds}")


if __name__ == "__main__":
    cli()



import ssl
from socket import socket, AF_INET, SOCK_STREAM
from binascii import hexlify
import idna
import validators
from certifi import where
from OpenSSL import SSL
from OpenSSL.crypto import X509, FILETYPE_PEM, load_certificate
from cryptography import x509
from cryptography.x509.base import Certificate
from cryptography.x509.extensions import (
    Extension,
    SubjectKeyIdentifier,
    AuthorityKeyIdentifier,
)
from .context import *  # noqa: F403
from .stores import VERSIONS
from .stores.android_2_2 import PEM_FILES as ANDROID2_2_PEM_FILES
from .stores.android_2_3 import PEM_FILES as ANDROID2_3_PEM_FILES
from .stores.android_3 import PEM_FILES as ANDROID3_PEM_FILES
from .stores.android_4_4 import PEM_FILES as ANDROID4_4_PEM_FILES
from .stores.android_4 import PEM_FILES as ANDROID4_PEM_FILES
from .stores.android_7 import PEM_FILES as ANDROID7_PEM_FILES
from .stores.android_8 import PEM_FILES as ANDROID8_PEM_FILES
from .stores.android_9 import PEM_FILES as ANDROID9_PEM_FILES
from .stores.android_10 import PEM_FILES as ANDROID10_PEM_FILES
from .stores.android_11 import PEM_FILES as ANDROID11_PEM_FILES
from .stores.android_12 import PEM_FILES as ANDROID12_PEM_FILES
from .stores.android_13 import PEM_FILES as ANDROID13_PEM_FILES
from .stores.android_14 import PEM_FILES as ANDROID14_PEM_FILES
from .stores.android_latest import PEM_FILES as ANDROID_PEM_FILES
from .stores.ccadb import PEM_FILES as CCADB_PEM_FILES
from .stores.java import PEM_FILES as JAVA_PEM_FILES
from .stores.linux import PEM_FILES as LINUX_PEM_FILES
from .stores.certifi import PEM_FILES as CERTIFI_PEM_FILES
from .stores.mintsifry_rossii import PEM_FILES as RUSSIA_PEM_FILES
from .stores.rustls import PEM_FILES as RUST_PEM_FILES

__module__ = "tlstrust.util"

MISSING_MESSAGE = "Certificate does not exist"


class InvalidChainError(ValueError):
    """Raised when the certificate chain is empty or missing a server leaf certificate"""


def valid_context_type(context_type: int) -> bool:
    contexts = {**SOURCES, **PLATFORMS, **BROWSERS, **LANGUAGES}
    return context_type is None or context_type in [contexts[x] for x in contexts]


def get_key_identifier_hex(cert: Certificate, extension: Extension, key: str) -> str:
    for ext in cert.extensions:
        if isinstance(ext.value, extension):
            return hexlify(getattr(ext.value, key)).decode("utf-8")


def match_certificate(aki, root_ca: X509) -> bool:
    return any(
        isinstance(ext.value, SubjectKeyIdentifier)
        and aki == hexlify(ext.value.key_identifier).decode("utf-8")
        for ext in root_ca.to_cryptography().extensions
    )


def get_certificate_chain(
    host: str, port: int, use_sni: bool = True, client_cert: X509 = None
) -> tuple[list[X509], str]:
    if not isinstance(port, int):
        raise TypeError(f"provided an invalid type {type(port)} for port, expected int")
    if validators.domain(host) is not True:
        raise ValueError(f"provided an invalid domain {host}")

    for method in [
        "TLSv1_METHOD",
        "TLSv1_1_METHOD",
        "TLSv1_2_METHOD",
        "SSLv23_METHOD",
    ]:
        ctx = SSL.Context(method=getattr(SSL, method))
        ctx.load_verify_locations(cafile=where())
        if isinstance(client_cert, X509):
            ctx.use_certificate(client_cert)
        ctx.verify_mode = SSL.VERIFY_NONE
        ctx.check_hostname = False
        sock = socket(AF_INET, SOCK_STREAM)
        sock.settimeout(2)
        conn = SSL.Connection(ctx, sock)
        if all([ssl.HAS_SNI, use_sni]):
            conn.set_tlsext_host_name(idna.encode(host))
        certificate_chain = []
        skip = False
        try:
            conn.connect((host, port))
            conn.set_connect_state()
            conn.setblocking(1)
            conn.do_handshake()
            peer_address, _ = conn.getpeername()
            for (_, cert) in enumerate(conn.get_peer_cert_chain()):
                certificate_chain.append(cert)
            conn.shutdown()
        except SSL.Error:
            skip = True
        finally:
            conn.close()
        if skip:
            continue
        return certificate_chain, peer_address


def get_certificate_from_store(aki, context_type: int) -> X509:
    if not valid_context_type(context_type):
        raise AttributeError(INVALID_CONTEXT.format(context_type))
    certificate = None
    try:
        if context_type == SOURCE_CCADB:
            certificate = load_certificate(FILETYPE_PEM, CCADB_PEM_FILES[aki].encode())
        if context_type == SOURCE_ANDROID:
            certificate = load_certificate(
                FILETYPE_PEM, ANDROID_PEM_FILES[aki].encode()
            )
        if context_type == SOURCE_JAVA:
            certificate = load_certificate(FILETYPE_PEM, JAVA_PEM_FILES[aki].encode())
        if context_type == SOURCE_LINUX:
            certificate = load_certificate(FILETYPE_PEM, LINUX_PEM_FILES[aki].encode())
        if context_type == SOURCE_RUSSIA:
            certificate = load_certificate(FILETYPE_PEM, RUSSIA_PEM_FILES[aki].encode())
        if context_type == SOURCE_RUSTLS:
            certificate = load_certificate(FILETYPE_PEM, RUST_PEM_FILES[aki].encode())
        if context_type == SOURCE_CERTIFI:
            certificate = load_certificate(
                FILETYPE_PEM, CERTIFI_PEM_FILES[aki].encode()
            )
        if context_type == PLATFORM_ANDROID2_2:
            certificate = load_certificate(
                FILETYPE_PEM, ANDROID2_2_PEM_FILES[aki].encode()
            )
        if context_type == PLATFORM_ANDROID2_3:
            certificate = load_certificate(
                FILETYPE_PEM, ANDROID2_3_PEM_FILES[aki].encode()
            )
        if context_type == PLATFORM_ANDROID3:
            certificate = load_certificate(
                FILETYPE_PEM, ANDROID3_PEM_FILES[aki].encode()
            )
        if context_type == PLATFORM_ANDROID4:
            certificate = load_certificate(
                FILETYPE_PEM, ANDROID4_PEM_FILES[aki].encode()
            )
        if context_type == PLATFORM_ANDROID4_4:
            certificate = load_certificate(
                FILETYPE_PEM, ANDROID4_4_PEM_FILES[aki].encode()
            )
        if context_type == PLATFORM_ANDROID7:
            certificate = load_certificate(
                FILETYPE_PEM, ANDROID7_PEM_FILES[aki].encode()
            )
        if context_type == PLATFORM_ANDROID8:
            certificate = load_certificate(
                FILETYPE_PEM, ANDROID8_PEM_FILES[aki].encode()
            )
        if context_type == PLATFORM_ANDROID9:
            certificate = load_certificate(
                FILETYPE_PEM, ANDROID9_PEM_FILES[aki].encode()
            )
        if context_type == PLATFORM_ANDROID10:
            certificate = load_certificate(
                FILETYPE_PEM, ANDROID10_PEM_FILES[aki].encode()
            )
        if context_type == PLATFORM_ANDROID11:
            certificate = load_certificate(
                FILETYPE_PEM, ANDROID11_PEM_FILES[aki].encode()
            )
        if context_type == PLATFORM_ANDROID12:
            certificate = load_certificate(
                FILETYPE_PEM, ANDROID12_PEM_FILES[aki].encode()
            )
        if context_type == PLATFORM_ANDROID13:
            certificate = load_certificate(
                FILETYPE_PEM, ANDROID13_PEM_FILES[aki].encode()
            )
        if context_type == PLATFORM_ANDROID14:
            certificate = load_certificate(
                FILETYPE_PEM, ANDROID14_PEM_FILES[aki].encode()
            )
    except KeyError:
        pass
    if certificate is None or not match_certificate(aki, certificate):
        raise FileExistsError(MISSING_MESSAGE)
    return certificate


def get_cn_or_org(certificate: X509) -> str:
    cn_oid = certificate.to_cryptography().subject.get_attributes_for_oid(
        x509.OID_COMMON_NAME
    )
    if cn_oid:
        name = cn_oid[0]._value  # pylint: disable=protected-access
    else:
        name = (
            certificate.to_cryptography()
            .subject.get_attributes_for_oid(x509.OID_ORGANIZATION_NAME)[0]
            ._value
        )  # pylint: disable=protected-access
    return name


def get_leaf(certificates: list[X509]) -> X509:
    new_chain = []
    akis_paris = []
    for cert in certificates:
        common_name = get_cn_or_org(cert)
        aki = get_key_identifier_hex(
            cert.to_cryptography(),
            extension=AuthorityKeyIdentifier,
            key="key_identifier",
        )
        akis_paris.append((aki, cert))
        if common_name[0:1] == "*" or validators.domain(common_name):
            new_chain.append(cert)
    if len(new_chain) == 1:
        return new_chain[0]


def build_chains(leaf: X509, certificates: list[X509]) -> dict:
    roots: list[X509] = []
    chains = {}
    leaf_aki = get_key_identifier_hex(
        leaf.to_cryptography(), extension=AuthorityKeyIdentifier, key="key_identifier"
    )
    aki_lookup = {leaf_aki: [leaf]}
    for cert in certificates:
        aki = get_key_identifier_hex(
            cert.to_cryptography(),
            extension=AuthorityKeyIdentifier,
            key="key_identifier",
        )
        if aki == leaf_aki:
            continue
        aki_lookup.setdefault(aki, [])
        aki_lookup[aki].append(cert)
        for _, context_type in STORES.items():
            try:
                ret = get_certificate_from_store(aki, context_type)
            except FileExistsError:
                continue
            root_ski = get_key_identifier_hex(
                ret.to_cryptography(), extension=SubjectKeyIdentifier, key="digest"
            )
            if root_ski not in [
                get_key_identifier_hex(
                    c.to_cryptography(), extension=SubjectKeyIdentifier, key="digest"
                )
                for c in roots
            ]:
                roots.append(ret)

    def next_chain(ski: str, lookup: dict) -> list[dict]:
        chain = []
        for next_cert in lookup.get(ski, []):
            next_common_name = get_cn_or_org(next_cert)
            next_ski = get_key_identifier_hex(
                next_cert.to_cryptography(),
                extension=SubjectKeyIdentifier,
                key="digest",
            )
            next_aki = get_key_identifier_hex(
                next_cert.to_cryptography(),
                extension=AuthorityKeyIdentifier,
                key="key_identifier",
            )
            chain.append(
                {
                    "certificate": next_cert,
                    "ski": next_ski,
                    "aki": next_aki,
                    "common_name": next_common_name,
                    "next": next_chain(next_ski, lookup),
                }
            )
        return chain

    index = 0
    for cert in roots:
        ski = get_key_identifier_hex(
            cert.to_cryptography(), extension=SubjectKeyIdentifier, key="digest"
        )
        common_name = get_cn_or_org(cert)
        chains[str(index)] = {
            "certificate": cert,
            "ski": ski,
            "common_name": common_name,
            "next": next_chain(ski, aki_lookup),
        }
        index += 1
    return chains


def get_store_result_text(name: str, **kwargs) -> dict:
    short_name = SHORT_LOOKUP[name]
    trust_status = f"No Root CA Certificate in the {short_name} Trust Store"
    if kwargs.get("exists"):
        trust_status = (
            f"Root CA Certificate present in {short_name} {VERSIONS[name]} Trust Store"
        )
        if name == CCADB:
            trust_status += " (Mozilla, Microsoft, and Apple)"
        if name == PYTHON_CERTIFI:
            trust_status += " (Django, requests, urllib, and anything based from these)"
        if kwargs.get("expired"):
            trust_status += " EXPIRED"

    return trust_status
